import random
from itertools import product
import future_uuid


def test_uuid8():
    u = future_uuid.uuid8()

    assert u.variant == future_uuid.uuid.RFC_4122
    assert u.version == 8

    for _, hi, mid, lo in product(
        range(10),  # repeat 10 times
        [None, 0, random.getrandbits(48)],
        [None, 0, random.getrandbits(12)],
        [None, 0, random.getrandbits(62)],
    ):
        u = future_uuid.uuid8(hi, mid, lo)
        assert u.variant == future_uuid.uuid.RFC_4122
        assert u.version == 8
        if hi is not None:
            assert (u.int >> 80) & 0xFFFFFFFFFFFF == hi
        if mid is not None:
            assert (u.int >> 64) & 0xFFF == mid
        if lo is not None:
            assert u.int & 0x3FFFFFFFFFFFFFFF == lo


def test_uuid8_uniqueness():
    # Test that UUIDv8-generated values are unique
    # (up to a negligible probability of failure).
    u1 = future_uuid.uuid8()
    u2 = future_uuid.uuid8()
    assert u1.int != u2.int
    assert u1.version == u2.version



import uuid
import os

_last_timestamp_v7 = None
_last_counter_v7 = 0  # 42-bit counter

_RFC_4122_VERSION_6_FLAGS = (6 << 76) | (0x8000 << 48)
_RFC_4122_VERSION_7_FLAGS = (7 << 76) | (0x8000 << 48)
_RFC_4122_VERSION_8_FLAGS = (8 << 76) | (0x8000 << 48)

_UINT_128_MAX = (1 << 128) - 1


def _uuid7_get_counter_and_tail():
    rand = int.from_bytes(os.urandom(10), byteorder="big")
    # 42-bit counter with MSB set to 0
    counter = (rand >> 32) & 0x1FF_FFFF_FFFF
    # 32-bit random data
    tail = rand & 0xFFFF_FFFF
    return counter, tail


class UUID(uuid.UUID):
    @property
    def time(self):
        if self.version == 6:
            # time_hi (32) | time_mid (16) | ver (4) | time_lo (12) | ... (64)
            time_hi = self.int >> 96
            time_lo = (self.int >> 64) & 0x0FFF
            return time_hi << 28 | (self.time_mid << 12) | time_lo
        else:
            # time_lo (32) | time_mid (16) | ver (4) | time_hi (12) | ... (64)
            return super().time


def _uuid_from_int(value):
    """Create a UUID from an integer *value*. Internal use only."""
    assert 0 <= value <= _UINT_128_MAX, repr(value)
    self = object.__new__(UUID)
    object.__setattr__(self, "int", value)
    object.__setattr__(self, "is_safe", uuid.SafeUUID.unknown)
    return self


def uuid7():
    """Generate a UUID from a Unix timestamp in milliseconds and random bits.

    UUIDv7 objects feature monotonicity within a millisecond.
    """
    # --- 48 ---   -- 4 --   --- 12 ---   -- 2 --   --- 30 ---   - 32 -
    # unix_ts_ms | version | counter_hi | variant | counter_lo | random
    #
    # 'counter = counter_hi | counter_lo' is a 42-bit counter constructed
    # with Method 1 of RFC 9562, §6.2, and its MSB is set to 0.
    #
    # 'random' is a 32-bit random value regenerated for every new UUID.
    #
    # If multiple UUIDs are generated within the same millisecond, the LSB
    # of 'counter' is incremented by 1. When overflowing, the timestamp is
    # advanced and the counter is reset to a random 42-bit integer with MSB
    # set to 0.

    global _last_timestamp_v7
    global _last_counter_v7

    import time

    nanoseconds = time.time_ns()
    timestamp_ms = nanoseconds // 1_000_000

    if _last_timestamp_v7 is None or timestamp_ms > _last_timestamp_v7:
        counter, tail = _uuid7_get_counter_and_tail()
    else:
        if timestamp_ms < _last_timestamp_v7:
            timestamp_ms = _last_timestamp_v7 + 1
        # advance the 42-bit counter
        counter = _last_counter_v7 + 1
        if counter > 0x3FF_FFFF_FFFF:
            # advance the 48-bit timestamp
            timestamp_ms += 1
            counter, tail = _uuid7_get_counter_and_tail()
        else:
            # 32-bit random data
            tail = int.from_bytes(os.urandom(4), byteorder="big")

    unix_ts_ms = timestamp_ms & 0xFFFF_FFFF_FFFF
    counter_msbs = counter >> 30
    # keep 12 counter's MSBs and clear variant bits
    counter_hi = counter_msbs & 0x0FFF
    # keep 30 counter's LSBs and clear version bits
    counter_lo = counter & 0x3FFF_FFFF
    # ensure that the tail is always a 32-bit integer (by construction,
    # it is already the case, but future interfaces may allow the user
    # to specify the random tail)
    tail &= 0xFFFF_FFFF

    int_uuid_7 = unix_ts_ms << 80
    int_uuid_7 |= counter_hi << 64
    int_uuid_7 |= counter_lo << 32
    int_uuid_7 |= tail
    # by construction, the variant and version bits are already cleared
    int_uuid_7 |= _RFC_4122_VERSION_7_FLAGS
    res = _uuid_from_int(int_uuid_7)

    # defer global update until all computations are done
    _last_timestamp_v7 = timestamp_ms
    _last_counter_v7 = counter
    return res


_last_timestamp_v6 = None


def uuid6(node=None, clock_seq=None):
    """Similar to :func:`uuid1` but where fields are ordered differently
    for improved DB locality.

    More precisely, given a 60-bit timestamp value as specified for UUIDv1,
    for UUIDv6 the first 48 most significant bits are stored first, followed
    by the 4-bit version (same position), followed by the remaining 12 bits
    of the original 60-bit timestamp.
    """
    global _last_timestamp_v6
    import time

    nanoseconds = time.time_ns()
    # 0x01b21dd213814000 is the number of 100-ns intervals between the
    # UUID epoch 1582-10-15 00:00:00 and the Unix epoch 1970-01-01 00:00:00.
    timestamp = nanoseconds // 100 + 0x01B21DD213814000
    if _last_timestamp_v6 is not None and timestamp <= _last_timestamp_v6:
        timestamp = _last_timestamp_v6 + 1
    _last_timestamp_v6 = timestamp
    if clock_seq is None:
        import random

        clock_seq = random.getrandbits(14)  # instead of stable storage
    time_hi_and_mid = (timestamp >> 12) & 0xFFFF_FFFF_FFFF
    time_lo = timestamp & 0x0FFF  # keep 12 bits and clear version bits
    clock_s = clock_seq & 0x3FFF  # keep 14 bits and clear variant bits
    if node is None:
        node = uuid.getnode()
    # --- 32 + 16 ---   -- 4 --   -- 12 --  -- 2 --   -- 14 ---    48
    # time_hi_and_mid | version | time_lo | variant | clock_seq | node
    int_uuid_6 = time_hi_and_mid << 80
    int_uuid_6 |= time_lo << 64
    int_uuid_6 |= clock_s << 48
    int_uuid_6 |= node & 0xFFFF_FFFF_FFFF
    # by construction, the variant and version bits are already cleared
    int_uuid_6 |= _RFC_4122_VERSION_6_FLAGS
    return _uuid_from_int(int_uuid_6)


def uuid8(a=None, b=None, c=None):
    """Generate a UUID from three custom blocks.

    * 'a' is the first 48-bit chunk of the UUID (octets 0-5);
    * 'b' is the mid 12-bit chunk (octets 6-7);
    * 'c' is the last 62-bit chunk (octets 8-15).

    When a value is not specified, a pseudo-random value is generated.
    """
    if a is None:
        import random

        a = random.getrandbits(48)
    if b is None:
        import random

        b = random.getrandbits(12)
    if c is None:
        import random

        c = random.getrandbits(62)
    int_uuid_8 = (a & 0xFFFF_FFFF_FFFF) << 80
    int_uuid_8 |= (b & 0xFFF) << 64
    int_uuid_8 |= c & 0x3FFF_FFFF_FFFF_FFFF
    # by construction, the variant and version bits are already cleared
    int_uuid_8 |= _RFC_4122_VERSION_8_FLAGS
    return _uuid_from_int(int_uuid_8)



from unittest import mock
import random
import future_uuid
import uuid


def test_uuid6():
    u = future_uuid.uuid6()
    assert u.variant == future_uuid.uuid.RFC_4122
    assert u.version == 6

    fake_nanoseconds = 0x1571_20A1_DE1A_C533
    fake_node_value = 0x54E1_ACF6_DA7F
    fake_clock_seq = 0x14C5
    with (
        mock.patch.object(future_uuid, "_last_timestamp_v6", None),
        mock.patch.object(uuid, "getnode", return_value=fake_node_value),
        mock.patch("time.time_ns", return_value=fake_nanoseconds),
        mock.patch("random.getrandbits", return_value=fake_clock_seq),
    ):
        u = future_uuid.uuid6()
        assert u.variant == future_uuid.uuid.RFC_4122
        assert u.version == 6

        # 32 (top) | 16 (mid) | 12 (low) == 60 (timestamp)
        assert u.time == 0x1E901FCA_7A55_B92
        assert u.fields[0] == 0x1E901FCA  # 32 top bits of time
        assert u.fields[1] == 0x7A55  # 16 mid bits of time
        # 4 bits of version + 12 low bits of time
        assert (u.fields[2] >> 12) & 0xF == 6
        assert (u.fields[2] & 0xFFF) == 0xB92
        # 2 bits of variant + 6 high bits of clock_seq
        assert (u.fields[3] >> 6) & 0xF == 2
        assert u.fields[3] & 0x3F == fake_clock_seq >> 8
        # 8 low bits of clock_seq
        assert u.fields[4] == fake_clock_seq & 0xFF
        assert u.fields[5] == fake_node_value


def test_uuid6_uniqueness():
    # Test that UUIDv6-generated values are unique.

    # Unlike UUIDv8, only 62 bits can be randomized for UUIDv6.
    # In practice, however, it remains unlikely to generate two
    # identical UUIDs for the same 60-bit timestamp if neither
    # the node ID nor the clock sequence is specified.
    uuids = {future_uuid.uuid6() for _ in range(1000)}
    assert len(uuids) == 1000
    versions = {u.version for u in uuids}
    assert versions == {6}

    timestamp = 0x1EC9414C_232A_B00
    fake_nanoseconds = (timestamp - 0x1B21DD21_3814_000) * 100

    with mock.patch("time.time_ns", return_value=fake_nanoseconds):

        def gen():
            with mock.patch.object(future_uuid, "_last_timestamp_v6", None):
                return future_uuid.uuid6(node=0, clock_seq=None)

        # By the birthday paradox, sampling N = 1024 UUIDs with identical
        # node IDs and timestamps results in duplicates with probability
        # close to 1 (not having a duplicate happens with probability of
        # order 1E-15) since only the 14-bit clock sequence is randomized.
        N = 1024
        uuids = {gen() for _ in range(N)}
        assert {u.node for u in uuids} == {0}
        assert {u.time for u in uuids} == {timestamp}
        assert len(uuids) < N, "collision property does not hold"


def test_uuid6_node():
    # Make sure the given node ID appears in the UUID.
    #
    # Note: when no node ID is specified, the same logic as for UUIDv1
    # is applied to UUIDv6. In particular, there is no need to test that
    # getnode() correctly returns positive integers of exactly 48 bits
    # since this is done in test_uuid1_eui64().
    assert future_uuid.uuid6().node.bit_length() <= 48

    assert future_uuid.uuid6(0).node == 0

    # tests with explicit values
    max_node = 0xFFFF_FFFF_FFFF
    assert future_uuid.uuid6(max_node).node == max_node
    big_node = 0xE_1234_5678_ABCD  # 52-bit node
    res_node = 0x0_1234_5678_ABCD  # truncated to 48 bits
    assert future_uuid.uuid6(big_node).node == res_node

    # randomized tests
    for _ in range(10):
        # node with > 48 bits is truncated
        for b in [24, 48, 72]:
            node = (1 << (b - 1)) | random.getrandbits(b)
            assert node.bit_length() == b
            u = future_uuid.uuid6(node=node)
            assert u.node == node & 0xFFFF_FFFF_FFFF


def test_uuid6_clock_seq():
    # Make sure the supplied clock sequence appears in the UUID.
    #
    # For UUIDv6, clock sequence bits are stored from bit 48 to bit 62,
    # with the convention that the least significant bit is bit 0 and
    # the most significant bit is bit 127.
    def get_clock_seq(u):
        return (u.int >> 48) & 0x3FFF

    u = future_uuid.uuid6()
    assert get_clock_seq(u).bit_length() <= 14

    # tests with explicit values
    big_clock_seq = 0xFFFF  # 16-bit clock sequence
    res_clock_seq = 0x3FFF  # truncated to 14 bits
    u = future_uuid.uuid6(clock_seq=big_clock_seq)
    assert get_clock_seq(u) == res_clock_seq

    # some randomized tests
    for _ in range(10):
        # clock_seq with > 14 bits is truncated
        for b in [7, 14, 28]:
            node = random.getrandbits(48)
            clock_seq = (1 << (b - 1)) | random.getrandbits(b)
            assert clock_seq.bit_length() == b
            u = future_uuid.uuid6(node=node, clock_seq=clock_seq)
            assert get_clock_seq(u) == clock_seq & 0x3FFF


def test_uuid6_test_vectors():
    # https://www.rfc-editor.org/rfc/rfc9562#name-test-vectors
    # (separators are put at the 12th and 28th bits)
    timestamp = 0x1EC9414C_232A_B00
    fake_nanoseconds = (timestamp - 0x1B21DD21_3814_000) * 100
    # https://www.rfc-editor.org/rfc/rfc9562#name-example-of-a-uuidv6-value
    node = 0x9F6BDECED846
    clock_seq = (3 << 12) | 0x3C8

    with (
        mock.patch.object(future_uuid, "_last_timestamp_v6", None),
        mock.patch("time.time_ns", return_value=fake_nanoseconds),
    ):
        u = future_uuid.uuid6(node=node, clock_seq=clock_seq)
        assert str(u).upper() == "1EC9414C-232A-6B00-B3C8-9F6BDECED846"
        #   32          16      4      12       2      14         48
        # time_hi | time_mid | ver | time_lo | var | clock_seq | node
        assert u.time == timestamp
        assert u.int & 0xFFFF_FFFF_FFFF == node
        assert (u.int >> 48) & 0x3FFF == clock_seq
        assert (u.int >> 62) & 0x3 == 0b10
        assert (u.int >> 64) & 0xFFF == 0xB00
        assert (u.int >> 76) & 0xF == 0x6
        assert (u.int >> 80) & 0xFFFF == 0x232A
        assert (u.int >> 96) & 0xFFFF_FFFF == 0x1EC9_414C



from unittest import mock
import future_uuid
import random
import os


def test_uuid7():
    u = future_uuid.uuid7()
    assert u.variant == future_uuid.uuid.RFC_4122
    assert u.version == 7

    # 1 Jan 2023 12:34:56.123_456_789
    timestamp_ns = 1672533296_123_456_789  # ns precision
    timestamp_ms, _ = divmod(timestamp_ns, 1_000_000)

    for _ in range(100):
        counter_hi = random.getrandbits(11)
        counter_lo = random.getrandbits(30)
        counter = (counter_hi << 30) | counter_lo

        tail = random.getrandbits(32)
        # effective number of bits is 32 + 30 + 11 = 73
        random_bits = counter << 32 | tail

        # set all remaining MSB of fake random bits to 1 to ensure that
        # the implementation correctly removes them
        random_bits = (((1 << 7) - 1) << 73) | random_bits
        random_data = random_bits.to_bytes(10, byteorder="big")

        with (
            mock.patch.multiple(
                future_uuid,
                _last_timestamp_v7=None,
                _last_counter_v7=0,
            ),
            mock.patch("time.time_ns", return_value=timestamp_ns),
            mock.patch("os.urandom", return_value=random_data) as urand,
        ):
            u = future_uuid.uuid7()
            urand.assert_called_once_with(10)
            assert u.variant == future_uuid.uuid.RFC_4122
            assert u.version == 7

            assert future_uuid._last_timestamp_v7 == timestamp_ms
            assert future_uuid._last_counter_v7 == counter

            unix_ts_ms = timestamp_ms & 0xFFFF_FFFF_FFFF
            assert (u.int >> 80) & 0xFFFF_FFFF_FFFF == unix_ts_ms

            assert (u.int >> 75) & 1 == 0  # check that the MSB is 0
            assert (u.int >> 64) & 0xFFF == counter_hi
            assert (u.int >> 32) & 0x3FFF_FFFF == counter_lo
            assert u.int & 0xFFFF_FFFF == tail


def test_uuid7_uniqueness():
    # Test that UUIDv7-generated values are unique.
    #
    # While UUIDv8 has an entropy of 122 bits, those 122 bits may not
    # necessarily be sampled from a PRNG. On the other hand, UUIDv7
    # uses os.urandom() as a PRNG which features better randomness.
    N = 1000
    uuids = {future_uuid.uuid7() for _ in range(N)}
    assert len(uuids) == N

    versions = {u.version for u in uuids}
    assert versions == {7}


def test_uuid7_monotonicity():
    us = [future_uuid.uuid7() for _ in range(10_000)]
    assert us == sorted(us)

    with mock.patch.multiple(
        future_uuid,
        _last_timestamp_v7=0,
        _last_counter_v7=0,
    ):
        # 1 Jan 2023 12:34:56.123_456_789
        timestamp_ns = 1672533296_123_456_789  # ns precision
        timestamp_ms, _ = divmod(timestamp_ns, 1_000_000)

        # counter_{hi,lo} are chosen so that "counter + 1" does not overflow
        counter_hi = random.getrandbits(11)
        counter_lo = random.getrandbits(29)
        counter = (counter_hi << 30) | counter_lo
        assert counter + 1 < 0x3FF_FFFF_FFFF

        tail = random.getrandbits(32)
        random_bits = counter << 32 | tail
        random_data = random_bits.to_bytes(10, byteorder="big")

        with (
            mock.patch("time.time_ns", return_value=timestamp_ns),
            mock.patch("os.urandom", return_value=random_data) as urand,
        ):
            u1 = future_uuid.uuid7()
            urand.assert_called_once_with(10)
            assert future_uuid._last_timestamp_v7 == timestamp_ms
            assert future_uuid._last_counter_v7 == counter
            assert (u1.int >> 64) & 0xFFF == counter_hi
            assert (u1.int >> 32) & 0x3FFF_FFFF == counter_lo
            assert u1.int & 0xFFFF_FFFF == tail

        # 1 Jan 2023 12:34:56.123_457_032 (same millisecond but not same ns)
        next_timestamp_ns = 1672533296_123_457_032
        next_timestamp_ms, _ = divmod(timestamp_ns, 1_000_000)
        assert timestamp_ms == next_timestamp_ms

        next_tail_bytes = os.urandom(4)
        next_fail = int.from_bytes(next_tail_bytes, byteorder="big")

        with (
            mock.patch("time.time_ns", return_value=next_timestamp_ns),
            mock.patch("os.urandom", return_value=next_tail_bytes) as urand,
        ):
            u2 = future_uuid.uuid7()
            urand.assert_called_once_with(4)
            # same milli-second
            assert future_uuid._last_timestamp_v7 == timestamp_ms
            # 42-bit counter advanced by 1
            assert future_uuid._last_counter_v7 == counter + 1
            assert (u2.int >> 64) & 0xFFF == counter_hi
            assert (u2.int >> 32) & 0x3FFF_FFFF == counter_lo + 1
            assert u2.int & 0xFFFF_FFFF == next_fail

        assert u1 < u2


def test_uuid7_timestamp_backwards():
    # 1 Jan 2023 12:34:56.123_456_789
    timestamp_ns = 1672533296_123_456_789  # ns precision
    timestamp_ms, _ = divmod(timestamp_ns, 1_000_000)
    fake_last_timestamp_v7 = timestamp_ms + 1

    # counter_{hi,lo} are chosen so that "counter + 1" does not overflow
    counter_hi = random.getrandbits(11)
    counter_lo = random.getrandbits(29)
    counter = (counter_hi << 30) | counter_lo
    assert counter + 1 < 0x3FF_FFFF_FFFF

    tail_bytes = os.urandom(4)
    tail = int.from_bytes(tail_bytes, byteorder="big")

    with (
        mock.patch.multiple(
            future_uuid,
            _last_timestamp_v7=fake_last_timestamp_v7,
            _last_counter_v7=counter,
        ),
        mock.patch("time.time_ns", return_value=timestamp_ns),
        mock.patch("os.urandom", return_value=tail_bytes) as urand,
    ):
        u = future_uuid.uuid7()
        urand.assert_called_once_with(4)
        assert u.variant == future_uuid.uuid.RFC_4122
        assert u.version == 7
        assert future_uuid._last_timestamp_v7 == fake_last_timestamp_v7 + 1
        unix_ts_ms = (fake_last_timestamp_v7 + 1) & 0xFFFF_FFFF_FFFF
        assert (u.int >> 80) & 0xFFFF_FFFF_FFFF == unix_ts_ms
        # 42-bit counter advanced by 1
        assert future_uuid._last_counter_v7 == counter + 1
        assert (u.int >> 64) & 0xFFF == counter_hi
        # 42-bit counter advanced by 1 (counter_hi is untouched)
        assert (u.int >> 32) & 0x3FFF_FFFF == counter_lo + 1
        assert u.int & 0xFFFF_FFFF == tail


def test_uuid7_overflow_counter():
    # 1 Jan 2023 12:34:56.123_456_789
    timestamp_ns = 1672533296_123_456_789  # ns precision
    timestamp_ms, _ = divmod(timestamp_ns, 1_000_000)

    new_counter_hi = random.getrandbits(11)
    new_counter_lo = random.getrandbits(30)
    new_counter = (new_counter_hi << 30) | new_counter_lo

    tail = random.getrandbits(32)
    random_bits = (new_counter << 32) | tail
    random_data = random_bits.to_bytes(10, byteorder="big")

    with (
        mock.patch.multiple(
            future_uuid,
            _last_timestamp_v7=timestamp_ms,
            # same timestamp, but force an overflow on the counter
            _last_counter_v7=0x3FF_FFFF_FFFF,
        ),
        mock.patch("time.time_ns", return_value=timestamp_ns),
        mock.patch("os.urandom", return_value=random_data) as urand,
    ):
        u = future_uuid.uuid7()
        urand.assert_called_with(10)
        assert u.variant == future_uuid.uuid.RFC_4122
        assert u.version == 7
        # timestamp advanced due to overflow
        assert future_uuid._last_timestamp_v7 == timestamp_ms + 1
        unix_ts_ms = (timestamp_ms + 1) & 0xFFFF_FFFF_FFFF
        assert (u.int >> 80) & 0xFFFF_FFFF_FFFF == unix_ts_ms
        # counter overflowed, so we picked a new one
        assert future_uuid._last_counter_v7 == new_counter
        assert (u.int >> 64) & 0xFFF == new_counter_hi
        assert (u.int >> 32) & 0x3FFF_FFFF == new_counter_lo
        assert u.int & 0xFFFF_FFFF == tail
